#!/usr/bin/env perl -w

# 20210201

use strict;
use warnings;

# 2018/9/2 Jun Inoue

my @files = <DATA>;

my $count = 1;
for my $files (@files) {
  chomp($files);
  next if $files =~ /^$/;

  #my ($spNamePrefix, $infile_pep, $infile_DNA) = split(/ +/, $files);
  my ($infile_pep, $infile_DNA) = split(/ +/, $files);
  my $spNamePrefix = "XXX_xxx";
  
  unless ($infile_DNA) {
    print "Error: PEP and DNA file names should be separated by space at the bottom of 000_control.pl.\n";
    exit;
  }

  print "## $count ############\n";
  print "## $infile_pep\n";
  print "## $infile_DNA\n\n";

  my $tempPEP_010out_onlyLongestPEP = "TEMP_010out_OnlyLongestPEP_".$infile_pep;
  my $temp_020out_onlyLongestDNA    = "TEMP_020out_OnlyLongestCDS_".$infile_DNA;

  my $temp_040out_revisedPEP        = "REV_040out_".$infile_pep;



=pod     ###### Analysis for files derived from recCollectFromGBK (NCBI database)

  #### Analyzing 
  print "$count. ###### 0.Change name line ######\n";  # Require time

  print "\nAnalysis start for infile_pep: $infile_pep\n";
  &change_nameLine005("prot", $spNamePrefix, $infile_pep);

  print "\nAnalysis start for infile_DNA: $infile_DNA\n";
  &change_nameLine005("nucl", $spNamePrefix, $infile_DNA);

  #&change_nameLine007($infile_pep);

  print "\n\n";
  next;
  exit;
  ########
=cut    ###### End


  print "$count. ###### 1.Picking up the longest transcript (PEP) ######\n";  # Require time

  &longPepPicker010($spNamePrefix, $infile_pep, $tempPEP_010out_onlyLongestPEP);
  print "\n\n";

  print "$count. ###### 2.Picking up the corresponding DNA records ######\n";  # Require time
  &cDNApicker020($spNamePrefix, $tempPEP_010out_onlyLongestPEP, $infile_DNA, $temp_020out_onlyLongestDNA);
  print "\n\n";

  print "$count. ###### 3.Comparing selected AA and DNA using pal2nal ######\n";
  #print ("perl pal2nal14a.pl $tempPEP_010out_onlyLongestPEP $temp_020out_onlyLongestDNA > 030_out_1stPAL2NALcomparison.txt\n");
  system("perl pal2nal14a.pl $tempPEP_010out_onlyLongestPEP $temp_020out_onlyLongestDNA > 030_out_1stPAL2NALcomparison.txt");
  print "\n\n";

  print "$count. ###### 4.Revising PEP records ######\n";
  &revisingPEP040($spNamePrefix, "030_out_1stPAL2NALcomparison.txt", $tempPEP_010out_onlyLongestPEP, $temp_040out_revisedPEP);
  print "\n\n";

  print "$count. ###### 5.Confirmation: comparing AA and DNA using pal2nal ######\n";
  print ("perl pal2nal14a.pl $temp_040out_revisedPEP $temp_020out_onlyLongestDNA > 050_out_2ndPAL2NALcomparison.txt\n");
  system("perl pal2nal14a.pl $temp_040out_revisedPEP $temp_020out_onlyLongestDNA > 050_out_2ndPAL2NALcomparison.txt");
  print "\n\n";

  print "$count. ###### 6.formatdb for revised PEP file ######\n";
  my $finalPEP = "000_OnlyLongestPEP_".$infile_pep;
  my $finalDNA = "000_OnlyLongestCDS_".$infile_DNA;
  system("cp $temp_040out_revisedPEP $finalPEP");
  system("cp $temp_020out_onlyLongestDNA $finalDNA");
  #print ("formatdb -p T -i $finalPEP\n");
  #system("formatdb -p T -i $finalPEP");
  print ("./makeblastdb -dbtype prot -in $finalPEP\n");
  system("./makeblastdb -dbtype prot -in $finalPEP");
  print "\n\n";


  print "$count. ###### 7.Removing outfile files ######\n";
  system("rm $tempPEP_010out_onlyLongestPEP $temp_020out_onlyLongestDNA $temp_040out_revisedPEP 030_out_1stPAL2NALcomparison.txt 050_out_2ndPAL2NALcomparison.txt");
  print "\n\n";

  $count++;
}


#################### change_nameLine005 START
sub change_nameLine005
{
  my ($type_data, $spNamePrefix, $infile) = @_;

  #print "type_data $type_data\n";
  #exit;

  open(INFILE, "$infile") or die "$!";
  my @ins = <INFILE>;
  close INFILE;
  my ($ref_recs, $ref_nameLines) = &fasHashReader(\@ins);
  my @nameLines  = @$ref_nameLines;
  my %recs       = %$ref_recs;
  
  my $spNameLineModified = "";
  my $count = 0;
  open(OUT,">"."005_"."$infile");
  foreach my $spNameLine (@nameLines) {
    if ($type_data eq "prot"){
        $spNameLineModified = &renamer_prot($spNamePrefix, $spNameLine);
    } else {
        $spNameLineModified = &renamer_nucl( $spNamePrefix, $spNameLine);
    }
    print OUT "$spNameLineModified DBNLINE|$count|\n";
    print OUT "$recs{$spNameLine}\n";
    print "$count\n" if $count % 1000 == 0;
    $count++;
  }
  $count = $count - 1;
  close OUT;
}


sub change_nameLine007
{
  my ($infile) = @_;
  
  open(INFILE, "$infile") or die "$!";
  my @ins = <INFILE>;
  close INFILE;
  my ($ref_recs, $ref_nameLines) = &fasHashReader(\@ins);
  my @nameLines  = @$ref_nameLines;
  my %recs       = %$ref_recs;
  
  my $count = 0;
  open(OUT,">"."007_"."$infile");
  foreach my $spNameLine (@nameLines) {
    my $spNameLineModified = &renamer2($spNameLine);
    print OUT "$spNameLineModified\n";
    print OUT "$recs{$spNameLine}\n";
    print "$count\n" if $count % 100 == 0;
    $count++;
  }
  $count = $count - 1;
  close OUT;
}


#################### longPepPicker010 START
sub longPepPicker010
{
  my ($spNamePrefix, $infile_pep, $tempPEP_010out_onlyLongestPEP) = @_;
  
  open(INFILE, "$infile_pep") or die "$!";
  my @ins = <INFILE>;
  close INFILE;
  my ($ref_recs, $ref_nameLines) = &fasHashReader(\@ins);
  my @nameLines  = @$ref_nameLines;
  my %recs       = %$ref_recs;
  
  my ($ref_protIDs, $ref_cdnsIDs, $ref_geneIDs) = &idPicker($ref_nameLines);

  my @geneIDs = @$ref_geneIDs;
  my @protIDs = @$ref_protIDs;
  my @cdnsIDs = @$ref_cdnsIDs;
  
  print "records: ".@nameLines."\n";
  print "protIDs: ".@protIDs."\n";
  print "cdnaIDs: ".@cdnsIDs."\n";
  print "geneIDs: ".@geneIDs."\n";
  
  if (@protIDs == @geneIDs) {
      print "Same nums: protIDs and geneIDs\n";
      my $count = 0;
      open(OUT,">$tempPEP_010out_onlyLongestPEP");
      foreach my $spNameLine (@nameLines) {
        my $spNameLineModified = &renamer_prot($spNamePrefix, $spNameLine);
        print OUT "$spNameLineModified DBNLINE|$count|\n";
        my $sequence = $recs{$spNameLine};           
        $sequence =~ s/\*/X/g;                   ####################################    * => X
        print OUT "$sequence\n";
        print "$count\n" if $count % 100 == 0;
        $count++;
      }
      $count = $count - 1;
      close OUT;
  } else {
      print "Different nums: protIDs and geneIDs\n";
      my $count = 0;
      open(OUT,">$tempPEP_010out_onlyLongestPEP");
      foreach my $geneID (@geneIDs) {
        #print "geneID |$geneID|\n";
        #exit;
        my $geneIDspaced = $geneID." ";
        my @protIDsGreped = ();
        if(@protIDsGreped = grep(/gene:$geneIDspaced/, @nameLines)){
        } 
        elsif(@protIDsGreped = grep(/gene:$geneID$/, @nameLines)){
        } 
        else {
          print "Error: Cannot find $geneID in nameLines.\n";
          print "Stopped.\n";
          exit;
        }
    
        my $ref_longestRecord = &logestPicker(\@protIDsGreped, $ref_recs);
        my $spNameLine = $ref_longestRecord->[0];
        my $sequence   = $ref_longestRecord->[2];
    
        my $spNameLineModified = &renamer_prot($spNamePrefix, $spNameLine);
      
        print OUT "$spNameLineModified DBNLINE|$count|\n";
        $sequence =~ s/\*/X/g;                   ####################################    * => X
        print OUT "$sequence\n";
    
        print "$count\n" if $count % 100 == 0;
        $count++;
    
      }
      $count = $count - 1;
      close OUT;
  }

}


sub logestPicker {
  my ($ref_transcriptsInSameGeneID, $ref_recs) = @_;
  my @transcriptsInSameGeneID = @$ref_transcriptsInSameGeneID;
  my %recsSR       = %$ref_recs;
 
  my @recs2D = ( );
  foreach my $nameLine (@transcriptsInSameGeneID) {
    my $seq = $recsSR{$nameLine};                      
    push(@recs2D, [$nameLine, length($seq), $seq]);
  }
  my @recs2D2 = sort {$b->[1] <=> $a->[1]} @recs2D;
  my $ref_longestRecord = shift@recs2D2;
  return($ref_longestRecord);
}


sub idPicker {
  my ($ref_nameLinesSR) = @_;
  my @nameLinesSR = @$ref_nameLinesSR;
  

  my @protIDsSR = ( ); my @cdnaIDsSR = ( ); my @geneIDsSR = ( );   

  foreach my $nameline (@nameLinesSR) {
    unless($nameline =~ />([^ ]+) .*gene:([^ ]+) .*transcript:([^ ]+) / || $nameline =~ />([^ ]+) .*gene:([^ ]+) .*transcript:([^ ]+)$/){
      print "Error: Check the name line.\n\n";
      print "Follow the Ensembl style including description:\n";
      #print ">ENSP00000482894 pep:known chromosome:GRCh38:X:100632063:100637104:-1 gene:ENSG00000000003 transcript:ENST00000614008 gene_biot.....\n";
      print ">ENSP00000362111.4 pep chromosome:GRCh38:X:100628670:100636806:-1 gene:ENSG00000000003.14 transcript:ENST00000373020.8 gene_biotype:protein_coding transcript_biotype:protein_coding gene_symbol:TSPAN6 description:tetraspanin 6 [Source:HGNC Symbol;Acc:HGNC:11858]\n";
      print "The name line should include words, gene and transcript.\n\n";
      print "Your name line:\n";
      print "$nameline";
      exit;
    }
    my $protIDsSR = $1;
    my $geneIDsSR = $2;
    my $cdnaIDsSR = $3;
    #print "|$protIDsSR| |$cdnaIDsSR| |$geneIDsSR|\n";
    push(@protIDsSR, $protIDsSR);
    push(@cdnaIDsSR, $cdnaIDsSR);
    push(@geneIDsSR, $geneIDsSR);
  }

  my @protIDsUniq = sort(uniqArray(\@protIDsSR));
  my @cdnaIDsUniq = sort(uniqArray(\@cdnaIDsSR));
  my @geneIDsUniq = sort(uniqArray(\@geneIDsSR));
  
  return (\@protIDsUniq, \@cdnaIDsUniq, \@geneIDsUniq);
}


sub uniqArray{
  my $array = shift;
  my %hash  = ();

  foreach my $value (@$array ){
    #print "value|$value|\n";
    $hash{$value} = 1;
  }

  return(keys %hash);
}


sub fasHashReader
{
  my $name = "";
  my $seq = "";
  my $flag = 0;
  my @nameArraySR = ( );
  my %fasHashSR = ( );

  my ($inFileContSR) = @_;

  foreach my $inFileLine (@$inFileContSR) {
    chomp($inFileLine);
    if($inFileLine =~ /^>/){  
      if($flag == 0){
        $name = $inFileLine;
        $flag = 1;  
      } else {
        $fasHashSR{$name} = $seq;  
        push(@nameArraySR,$name);
        $name = "$inFileLine";
        $seq = "";
      }
    } else {
      $seq .= $inFileLine; 
    }
  }
  $fasHashSR{$name} = $seq;
  push(@nameArraySR,$name);

  return (\%fasHashSR,\@nameArraySR,);
}


sub renamer_prot
{
  my ($spNamePrefix, $spName_LineSR) = @_;
  #print "spName_LineSR $spName_LineSR\n";
  
  my $nuclIDsr = "";
  if ($spName_LineSR =~ /transcript:/) {
    $spName_LineSR =~ / transcript:([^ ]+) /;
    $nuclIDsr = $1;
  } 
  elsif ($spName_LineSR =~ />[^ ]+ type:/) {
    $spName_LineSR =~ />([^ ]+) type:/;
    $nuclIDsr = $1;
  }
  elsif ($spName_LineSR =~ />[^ ]+ cds /) {
    $spName_LineSR =~ />([^ ]+) cds /;
    $nuclIDsr = $1;
  }
  elsif ($spName_LineSR =~ />[^ ]+ gene=/) {
    $spName_LineSR =~ />([^ ]+) gene=/;
    $nuclIDsr = $1;
  }
  elsif ($spName_LineSR =~ />[^ ]+ gene[^ ]+ $/) {
    $spName_LineSR =~ />([^ ]+) gene[^ ]+ $/;
    $nuclIDsr = $1;
  }
  elsif ($spName_LineSR =~ />([^ ]+)$/) {
    $spName_LineSR =~ />([^ ]+)$/;
    $nuclIDsr = $1;
  }
  else {
    print 'Cannot find "transcript: ", "  type:", " cds ". Stopped.'."\n";
    exit;
  }
  $nuclIDsr =~ s/_/-/g;      # <======================================= Change "_" in nuclID
  #print "spName_LineSR $spName_LineSR\n";
  #print "nuclIDsr $nuclIDsr\n";
  #exit;


  my $description = "";
  if ($spName_LineSR =~ /description:/) {
      $spName_LineSR =~ /description:(.*$)/;
      $description = $1;

      $description =~ s/\[[^\]]+\]//g;

      $description =~ s/[\(\)]//g;
      $description =~ s/ $//;
      $description =~ s/[, \/;:+<>\'\|]/-/g;
      $description =~ s/-{2,}/-/g;
      $description =~ s/-$//;
  } else {
      $description = "NONE";
  }

  my $gene_symbol = "";
  if ($spName_LineSR =~ /gene_symbol:/) {
      #print "$spName_LineSR\n";
      $spName_LineSR =~ /gene_symbol:([^ ]+)/;
      $gene_symbol = $1;

      $gene_symbol =~ s/\[[^\]]+\]//g;

      $gene_symbol =~ s/[\(\)]//g;
      $gene_symbol =~ s/ $//;
      $gene_symbol =~ s/[, \/;:+<>\'\|]/-/g;
      $gene_symbol =~ s/-{2,}/-/g;
      $gene_symbol =~ s/-$//;
  } else {
      $gene_symbol = "NONE";
  }


  my $protID_function = "";
  if ($gene_symbol eq "NONE"){
      #$protID_function = ">".$spNamePrefix.$nuclIDsr."_".$description;
      $protID_function = ">".$nuclIDsr."_".$description;
  } else {
      #$protID_function = ">".$spNamePrefix.$nuclIDsr."_".$gene_symbol."-".$description;
      $protID_function = ">".$nuclIDsr."_".$gene_symbol."-".$description;
  }
  $protID_function =~ s/(.{60}).*$/$1/g;
  #print "protID_function $protID_function\n";
  #exit;

  $spName_LineSR =~ s/^>//;
  my $newNameLine = $protID_function." ".$spName_LineSR;
  #print "newNameLine $newNameLine\n";
  #exit;
  
  return ($newNameLine)
}


sub renamer_nucl
{
  my ($spNamePrefix, $spName_LineSR) = @_;
  #print "spName_LineSR |$spName_LineSR|\n";
  #exit;

  if ($spName_LineSR !~ />[^ ]+ gene:[^ ]+ /) {
    print 'Error: cannot find "transcript: ", "  type:", " cds ". Stopped.'."\n";
    exit;
  }
  return ($spName_LineSR)
}


sub renamer2
{
  my ($spName_LineSR) = @_;
  
  my $former = "";
  my $latter = "";
  if ($spName_LineSR =~ />[^ ]+ .*$/) {
    $spName_LineSR =~ />([^ ]+)( .*$)/;
    $former = $1;
    $latter = $2;
  }
  else {
    print 'Cannot find ">spNameXXXX ". Stopped.'."\n";
    exit;
  }

  $former =~ s/\[[^\]]+\]//g;

  $former =~ s/[\(\)]//g;
  $former =~ s/ $//;
  $former =~ s/[, \/;:+<>\'\|]/-/g;
  $former =~ s/-{2,}/-/g;
  $former =~ s/-$//;

  return (">".$former.$latter)
}
#################### longPepPicker010 END

#################### longPepPicker020 START
sub cDNApicker020
{
  my ($spNamePrefix, $tempPEP_010out_onlyLongestPEP, $infile_DNA, $temp_020out_onlyLongestDNA) = @_;
  my $cDNAid = "";
  my $countCDNAprint = 0;
  
  #print "pep db: $tempPEP_010out_onlyLongestPEP\n";
  #print "cdna db: $infile_DNA\n";
  
  open(AASEQ,"$tempPEP_010out_onlyLongestPEP") or die"$!";
  my @aaSeqFile = <AASEQ>;
  close AASEQ;
  my ($ref_fasHashAA,$ref_nameArrayAA) = &fasHashReader(\@aaSeqFile);
  
  open(DNASEQ,"$infile_DNA") or die"$!";
  my @dnaSeqFile = <DNASEQ>;
  close DNASEQ;
  my ($ref_fasHashDNA,$ref_nameArrayDNA) = &fasHashReader(\@dnaSeqFile);
  
  open(OUTA,">$temp_020out_onlyLongestDNA");
  my $count = 0;
  foreach my $aaNameLine (@$ref_nameArrayAA){
    #print "aaNameLine $aaNameLine\n";
    $aaNameLine =~ /transcript:([^ ]+)/;         # Pick up the transcript ID from the name line of AA file
    $cDNAid = $1;
    print "$count\n" if $count%100 == 0;

    ##  The following line may be changed according to the species database.
    my $cDNAidS = $cDNAid." ";
    my $cDNAnameLine = "";
    my @temp = ();
    if(@temp = grep(/$cDNAidS/,@$ref_nameArrayDNA)){
      $cDNAnameLine = shift(@temp);
    }
    elsif(@temp = grep(/$cDNAid$/,@$ref_nameArrayDNA)){
      $cDNAnameLine = shift(@temp);
    }
    else {
      print "Cannot find |$cDNAid| in $infile_DNA. Check the namelin of DNA record\n";
      exit;
    }
    my $cDNAnameLineTMP = $cDNAnameLine;

    #$cDNAnameLineTMP =~ s/>/>$spNamePrefix/g;   #### Species name attaching

    print OUTA "$cDNAnameLineTMP DBNLINE|$count|\n";
    print OUTA "$ref_fasHashDNA->{$cDNAnameLine}\n";
    $countCDNAprint++;
    $count++;
  }
  close OUTA;
  
  $count = $count - 1;

  print "pep : ".@$ref_nameArrayAA." records\n";
  print "cdna: $countCDNAprint found\n";
}
#################### longPepPicker020 OUT

#################### revisingPEP040 START
sub revisingPEP040
{

  #my $spNamePrefix                    = $ARGV[0];  ## Human_
  #my $pal2nalErrorFileName      = $ARGV[1];  ## 030_out.txt
  #my $infile_tempOnlyLongestpep = $ARGV[2];  ## TEMP_OnlyLongestPEP_infile_pep.txt
  #my $outfile                   = $ARGV[3];  ## REV_infile_pep.txt

  my ($spNamePrefix, $out030_1stPAL2NALcomparison, $tempPEP_010out_onlyLongestPEP, $temp_040out_revisedPEP) = @_;
  
  #print "$spNamePrefix\n";
  #print "$out030_1stPAL2NALcomparison\n";
  #print "$tempPEP_010out_onlyLongestPEP\n";
  #print "$temp_040out_revisedPEP\n";
  
  open(OUT,"$out030_1stPAL2NALcomparison") or die "Cannot open $out030_1stPAL2NALcomparison.\n";
  my @pal2nalErrorFileCont = <OUT>;
  close OUT;
  my @pal2nalErrorProtIDs =  grep(/>/, @pal2nalErrorFileCont);
  #print "pal2nalErrorLines|@pal2nalErrorProtIDs|\n";
  
  open(INFILE,"$tempPEP_010out_onlyLongestPEP") or die "$!";
  my @onlyLongestPepFileLines = <INFILE>;
  close INFILE;
  my ($ref_recsOnlyLongestPEP, $ref_nameLinesOnlyLongestPEP) = &fasHashReader(\@onlyLongestPepFileLines);
  my @nameLines_onlyLongestPEP = @$ref_nameLinesOnlyLongestPEP;
  my %recs_onlyLongestPEP      = %$ref_recsOnlyLongestPEP;
  
  open(OUT, ">$temp_040out_revisedPEP");
  my $count = 1;
  foreach my $nameLine (@nameLines_onlyLongestPEP){
    $nameLine =~ /^>([^ ]+) .*$/;
    my $spName_protID_description = $1;
    
    if(grep/$spName_protID_description/, @pal2nalErrorProtIDs){
      if ($recs_onlyLongestPEP{$nameLine} =~ /U/){
        $recs_onlyLongestPEP{$nameLine} =~ s/U/*/g;
        print "  rec.$count. $spName_protID_description: U changed to * =>\t$spName_protID_description\n";
      }
      if ($recs_onlyLongestPEP{$nameLine} =~ /^X/){
        $recs_onlyLongestPEP{$nameLine} =~ s/^X+//;
        print "  rec.$count. $spName_protID_description: First X deleted =>\t$spName_protID_description\n";
      }
      else {
        $recs_onlyLongestPEP{$nameLine} =~ s/\w$//;    
        print "  rec.$count. $spName_protID_description: Last one residue deleted =>\t$spName_protID_description\n";
      }
    }
  
    #$spName_Line = &renamer_prot($spNamePrefix, $spName_Line);
  
    print OUT "$nameLine\n";
    print OUT "$recs_onlyLongestPEP{$nameLine}\n";

    print "  rec.$count.\n" if $count % 100 == 0;
    $count++;
  }
  close OUT;
}

#infile005_genbank_pep.txt  infile005_genebank_cds.txt

#################### revisingPEP040 OUT

__END__
infile_pep.txt infile_cds.txt
